@@ -1,42 +1,57 @@
 /* message.c -- Message manipulation/parsing
  * 
- * Copyright 1998 Carnegie Mellon University
- * 
- * No warranties, either expressed or implied, are made regarding the
- * operation, use, or results of the software.
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
- * Permission to use, copy, modify and distribute this software and its
- * documentation is hereby granted for non-commercial purposes only
- * provided that this copyright notice appears in all copies and in
- * supporting documentation.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
  *
- * Permission is also granted to Internet Service Providers and others
- * entities to use the software for internal purposes.
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
  *
- * The distribution, modification or sale of a product which uses or is
- * based on the software, in whole or in part, for commercial purposes or
- * benefits requires specific, additional permission from:
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
  *
- *  Office of Technology Transfer
- *  Carnegie Mellon University
- *  5000 Forbes Avenue
- *  Pittsburgh, PA  15213-3890
- *  (412) 268-4387, fax: (412) 268-7395
- *  tech-transfer@andrew.cmu.edu
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
  */
 
 /*
- * $Id: message.c,v 1.65 1999/03/29 20:23:25 tjs Exp $
+ * $Id: message.c,v 1.88.16.12 2003/07/01 14:33:53 ken3 Exp $
  */
 
+#include <config.h>
+
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <stdio.h>
 #include <ctype.h>
 #include <string.h>
-#include <time.h>
 #include <syslog.h>
 #include <sys/types.h>
 #include <sys/uio.h>
@@ -44,7 +59,7 @@
 #include <netinet/in.h>
 #include <stdlib.h>
 
-#include "sysexits.h"
+#include "exitcodes.h"
 #include "imap_err.h"
 #include "prot.h"
 #include "map.h"
@@ -54,7 +69,8 @@
 #include "charset.h"
 #include "util.h"
 #include "xmalloc.h"
-#include "config.h"
+#include "global.h"
+#include "retry.h"
 
 /* Message being parsed */
 struct msg {
@@ -157,7 +173,6 @@ static void message_parse_type P((char *hdr, struct body *body));
 static void message_parse_params P((char *hdr, struct param **paramp));
 static void message_fold_params P((struct param **paramp));
 static void message_parse_language P((char *hdr, struct param **paramp));
-static time_t message_parse_date P((char *hdr));
 static void message_parse_rfc822space P((char **s));
 
 static void message_parse_multipart P((struct msg *msg,
@@ -188,7 +203,7 @@ static void message_write_searchaddr P((struct ibuf *ibuf,
 					struct address *addrlist));
 
 static void message_ibuf_init P((struct ibuf *ibuf));
-static message_ibuf_ensure P((struct ibuf *ibuf, unsigned len));
+static int message_ibuf_ensure P((struct ibuf *ibuf, unsigned len));
 static void message_ibuf_iov P((struct iovec *iov, struct ibuf *ibuf));
 static void message_ibuf_free P((struct ibuf *ibuf));
 static void message_free_body P((struct body *body));
@@ -197,7 +212,7 @@ static void message_free_body P((struct body *body));
  * Copy a message of 'size' bytes from 'from' to 'to',
  * ensuring minimal RFC-822 compliance.
  *
- * Caller must have initialized config_* routines (with config_init) to read
+ * Caller must have initialized config_* routines (with cyrus_init) to read
  * imapd.conf before calling.
  */
 int
@@ -211,13 +226,13 @@ unsigned size;
     int r = 0;
     int n;
     int sawcr = 0, sawnl;
-    int reject8bit = config_getswitch("reject8bit", 0);
+    int reject8bit = config_getswitch(IMAPOPT_REJECT8BIT);
     int inheader = 1, blankline = 1;
 
     while (size) {
 	n = prot_read(from, buf, size > 4096 ? 4096 : size);
 	if (!n) {
-	    syslog(LOG_ERR, "IOERROR: reading message: %m");
+	    syslog(LOG_ERR, "IOERROR: reading message: unexpected end of file");
 	    return IMAP_IOERROR;
 	}
 
@@ -288,7 +303,7 @@ unsigned size;
 	   Unfortunately, we still need to look for the end of the string. */
 	for(p = (unsigned char*) buf; *p; p++);
 	
-	sawnl = (p[-1] == '\n');
+	sawnl = (p > (unsigned char *)buf) && (p[-1] == '\n');
     }
 }
 
@@ -312,7 +327,7 @@ struct index_record *message_index;
     if (fstat(fileno(infile), &sbuf) == -1) {
 	syslog(LOG_ERR, "IOERROR: fstat on new message in %s: %m",
 	       mailbox->name);
-	fatal("can't fstat message file", EX_OSFILE);
+	fatal("can't fstat message file", EC_OSFILE);
     }
     map_refresh(fileno(infile), 1, &msg_base, &msg_len, sbuf.st_size,
 		"new message", mailbox->name);
@@ -346,10 +361,11 @@ struct index_record *message_index;
     message_parse_body(&msg, mailbox->format, &body,
 		       DEFAULT_CONTENT_TYPE, (struct boundary *)0);
     
-    message_index->sentdate = message_parse_date(body.date);
+    message_index->sentdate = message_parse_date(body.date, 0);
     message_index->size = body.header_size + body.content_size;
     message_index->header_size = body.header_size;
     message_index->content_offset = body.content_offset;
+    message_index->content_lines = body.content_lines;
 
     message_index->cache_offset = lseek(mailbox->cache_fd, 0, SEEK_CUR);
     n = message_write_cache(mailbox->cache_fd, &body);
@@ -366,7 +382,7 @@ struct index_record *message_index;
 /*
  * Parse a body-part
  */
-static 
+static int
 message_parse_body(msg, format, body, defaultContentType, boundaries)
 struct msg *msg;
 int format;
@@ -375,10 +391,9 @@ char *defaultContentType;
 struct boundary *boundaries;
 {
     struct boundary newboundaries;
-    static struct body zerobody;
     int sawboundary;
 
-    *body = zerobody;
+    memset(body, 0, sizeof(struct body));
     newboundaries.id = 0;
 
     /* No passed-in boundary structure, create a new, empty one */
@@ -403,7 +418,7 @@ struct boundary *boundaries;
 	body->subpart = (struct body *)xmalloc(sizeof(struct body));
 
 	if (sawboundary) {
-	    *body->subpart = zerobody;
+	    memset(body->subpart, 0, sizeof(struct body));
 	    message_parse_type(DEFAULT_CONTENT_TYPE, body->subpart);
 	}
 	else {
@@ -429,6 +444,8 @@ struct boundary *boundaries;
 
     /* Free up boundary storage if necessary */
     if (newboundaries.id) free(newboundaries.id);
+
+    return 0;
 }
 
 /*
@@ -464,12 +481,11 @@ struct boundary *boundaries;
     /* Slurp up all of the headers into 'headers' */
     while (message_getline(next, left, msg) &&
 	   (next[-1] != '\n' ||
-	    (format == MAILBOX_FORMAT_NETNEWS ?
-	     (*next != '\n') : (*next != '\r' || next[1] != '\n')))) {
+	    (*next != '\r' || next[1] != '\n'))) {
 
 	if (next[-1] == '\n' && *next == '-' &&
 	    message_pendingboundary(next, boundaries->id, &boundaries->count)) {
-	    body->boundary_size = strlen(next)+(format==MAILBOX_FORMAT_NETNEWS);
+	    body->boundary_size = strlen(next);
 	    body->boundary_lines++;
 	    if (next - 1 > headers) {
 		body->boundary_size += 2;
@@ -487,14 +503,6 @@ struct boundary *boundaries;
 	left -= len;
 	next += len;
 
-	/* If reading netnews format, convert LF to CRLF */
-	if (format == MAILBOX_FORMAT_NETNEWS && next[-1] == '\n') {
-	    next[-1] = '\r';
-	    *next++ = '\n';
-	    *next = '\0';
-	    left--;
-	}
-
 	/* Allocate more header space if necessary */
 	if (left < 100) {
 	    len = next - headers;
@@ -505,14 +513,6 @@ struct boundary *boundaries;
 	}
     }
 
-    /* If reading netnews format, convert terminating LF to CRLF */
-    if (format == MAILBOX_FORMAT_NETNEWS) {
-	next = headers + strlen(headers);
-	next[-1] = '\r';
-	*next++ = '\n';
-	*next = '\0';
-    }
-    
     body->content_offset = msg->offset;
     body->header_size = strlen(headers+1);
 
@@ -680,7 +680,7 @@ message_parse_address(hdr, addrp)
 char *hdr;
 struct address **addrp;
 {
-    char *hdrend, hdrendchar;
+    char *hdrend, hdrendchar = '\0';
 
     /* Find end of header */
     hdrend = hdr;
@@ -720,7 +720,7 @@ char **hdrp;
     if (!hdr) return;
 
     /* Find end of encoding token */
-    for (p = hdr; *p && !isspace(*p) && *p != '('; p++) {
+    for (p = hdr; *p && !isspace((int) *p) && *p != '('; p++) {
 	if (*p < ' ' || strchr(TSPECIALS, *p)) return;
     }
     len = p - hdr;
@@ -731,10 +731,9 @@ char **hdrp;
 
     /* Save encoding token */
     *hdrp = xmalloc(len + 1);
-    strncpy(*hdrp, hdr, len);
-    (*hdrp)[len] = '\0';
+    strlcpy(*hdrp, hdr, len + 1);
     for (p = *hdrp; *p; p++) {
-	if (islower(*p)) *p = toupper(*p);
+	if (islower((int) *p)) *p = toupper((int) *p);
     }
 }
 	
@@ -770,18 +769,17 @@ char **hdrp;
     /* Save header value */
     len = hdrend - hdr;
     *hdrp = xmalloc(len + 1);
-    strncpy(*hdrp, hdr, len);
-    (*hdrp)[len] = '\0';
+    strlcpy(*hdrp, hdr, len + 1);
 
-    /* Un-fold header */
+    /* Un-fold header (overlapping buffers, use memmove) */
     hdrend = *hdrp;
-    while (hdrend = strchr(hdrend, '\n')) {
+    while ((hdrend = strchr(hdrend, '\n'))!=NULL) {
 	if (hdrend > *hdrp && hdrend[-1] == '\r') {
 	    hdrend--;
-	    strcpy(hdrend, hdrend+2);
+	    memmove(hdrend, hdrend+2, strlen(hdrend+2)+1);
 	}
 	else {
-	    strcpy(hdrend, hdrend+1);
+	    memmove(hdrend, hdrend+1, strlen(hdrend+1)+1);
 	}
     }
 }
@@ -841,7 +839,7 @@ struct body *body;
 
     /* Find end of type token */
     type = hdr;
-    for (; *hdr && !isspace(*hdr) && *hdr != '/' && *hdr != '('; hdr++) {
+    for (; *hdr && !isspace((int) *hdr) && *hdr != '/' && *hdr != '('; hdr++) {
 	if (*hdr < ' ' || strchr(TSPECIALS, *hdr)) return;
     }
     typelen = hdr - type;
@@ -859,7 +857,7 @@ struct body *body;
 
     /* Find end of subtype token */
     subtype = hdr;
-    for (; *hdr && !isspace(*hdr) && *hdr != ';' && *hdr != '('; hdr++) {
+    for (; *hdr && !isspace((int) *hdr) && *hdr != ';' && *hdr != '('; hdr++) {
 	if (*hdr < ' ' || strchr(TSPECIALS, *hdr)) return;
     }
     subtypelen = hdr - subtype;
@@ -872,16 +870,14 @@ struct body *body;
 
     /* Save content type & subtype */
     body->type = xmalloc(typelen + 1);
-    strncpy(body->type, type, typelen);
-    body->type[typelen] = '\0';
+    strlcpy(body->type, type, typelen + 1);
     for (p = body->type; *p; p++) {
-	if (islower(*p)) *p = toupper(*p);
+	if (islower((int) *p)) *p = toupper((int) *p);
     }
     body->subtype = xmalloc(subtypelen + 1);
-    strncpy(body->subtype, subtype, subtypelen);
-    body->subtype[subtypelen] = '\0';
+    strlcpy(body->subtype, subtype, subtypelen + 1);
     for (p = body->subtype; *p; p++) {
-	if (islower(*p)) *p = toupper(*p);
+	if (islower((int) *p)) *p = toupper((int) *p);
     }
 
     /* Parse parameter list */
@@ -912,7 +908,7 @@ struct body *body;
 
     /* Find end of disposition token */
     disposition = hdr;
-    for (; *hdr && !isspace(*hdr) && *hdr != ';' && *hdr != '('; hdr++) {
+    for (; *hdr && !isspace((int) *hdr) && *hdr != ';' && *hdr != '('; hdr++) {
 	if (*hdr < ' ' || strchr(TSPECIALS, *hdr)) return;
     }
     dispositionlen = hdr - disposition;
@@ -925,10 +921,10 @@ struct body *body;
 
     /* Save content disposition */
     body->disposition = xmalloc(dispositionlen + 1);
-    strncpy(body->disposition, disposition, dispositionlen);
-    body->disposition[dispositionlen] = '\0';
+    strlcpy(body->disposition, disposition, dispositionlen + 1);
+
     for (p = body->disposition; *p; p++) {
-	if (islower(*p)) *p = toupper(*p);
+	if (islower((int) *p)) *p = toupper((int) *p);
     }
 
     /* Parse parameter list */
@@ -947,7 +943,6 @@ char *hdr;
 struct param **paramp;
 {
     struct param *param;
-    static struct param zeroparam;
     char *attribute;
     int attributelen;
     char *value;
@@ -961,7 +956,7 @@ struct param **paramp;
 
 	/* Find end of attribute */
 	attribute = hdr;
-	for (; *hdr && !isspace(*hdr) && *hdr != '=' && *hdr != '('; hdr++) {
+	for (; *hdr && !isspace((int) *hdr) && *hdr != '=' && *hdr != '('; hdr++) {
 	    if (*hdr < ' ' || strchr(TSPECIALS, *hdr)) return;
 	}
 	attributelen = hdr - attribute;
@@ -995,7 +990,7 @@ struct param **paramp;
 	    if (!*hdr++) return;
 	}
 	else {
-	    for (; *hdr && !isspace(*hdr) && *hdr != ';' && *hdr != '('; hdr++) {
+	    for (; *hdr && !isspace((int) *hdr) && *hdr != ';' && *hdr != '('; hdr++) {
 		if (*hdr < ' ' || strchr(TSPECIALS, *hdr)) return;
 	    }
 	}
@@ -1009,12 +1004,12 @@ struct param **paramp;
 		  
 	/* Save attribute/value pair */
 	*paramp = param = (struct param *)xmalloc(sizeof(struct param));
-	*param = zeroparam;
+	memset(param, 0, sizeof(struct param));
 	param->attribute = xmalloc(attributelen + 1);
-	strncpy(param->attribute, attribute, attributelen);
-	param->attribute[attributelen] = '\0';
+	strlcpy(param->attribute, attribute, attributelen + 1);
+
 	for (p = param->attribute; *p; p++) {
-	    if (islower(*p)) *p = toupper(*p);
+	    if (islower((int) *p)) *p = toupper((int) *p);
 	}
 	param->value = xmalloc(valuelen + 1);
 	if (*value == '\"') {
@@ -1028,8 +1023,7 @@ struct param **paramp;
 	    *p = '\0';
 	}
 	else {
-	    strncpy(param->value, value, valuelen);
-	    param->value[valuelen] = '\0';
+	    strlcpy(param->value, value, valuelen + 1);
 	}
 
 	/* Get ready to parse the next parameter */
@@ -1137,7 +1131,8 @@ message_fold_params(struct param **params)
 				     strlen((*continuation)->value) + 1);
 			from = (*continuation)->value;
 			to = thisparam->value + strlen(thisparam->value);
-			while (*to++ = *from++);
+			while ((*to++ = *from++)!= 0)
+			    { }
 		    }
 		}
 		else {
@@ -1149,7 +1144,8 @@ message_fold_params(struct param **params)
 				     strlen((*continuation)->value) + 1);
 			from = (*continuation)->value;
 			to = thisparam->value + strlen(thisparam->value);
-			while (*to++ = *from++);
+			while ((*to++ = *from++) != 0)
+			    { }
 		    }
 		    else {
 			/* Have to re-encode thisparam value */
@@ -1174,7 +1170,9 @@ message_fold_params(struct param **params)
 			    from++;
 			}
 			from = (*continuation)->value;
-			while (*to++ = *from++);
+
+			while ((*to++ = *from++)!=0)
+			    { }
 
 			free(thisparam->value);
 			thisparam->value = tmpvalue;
@@ -1212,7 +1210,6 @@ char *hdr;
 struct param **paramp;
 {
     struct param *param;
-    static struct param zeroparam;
     char *value;
     int valuelen;
     char *p;
@@ -1228,8 +1225,8 @@ struct param **paramp;
 
 	/* Find end of value */
 	value = hdr;
-	for (; *hdr && !isspace(*hdr) && *hdr != ',' && *hdr != '('; hdr++) {
-	    if (*hdr != '-' && !isalpha(*hdr)) return;
+	for (; *hdr && !isspace((int) *hdr) && *hdr != ',' && *hdr != '('; hdr++) {
+	    if (*hdr != '-' && !isalpha(((int) *hdr))) return;
 	}
 	valuelen = hdr - value;
 
@@ -1241,12 +1238,12 @@ struct param **paramp;
 		  
 	/* Save value pair */
 	*paramp = param = (struct param *)xmalloc(sizeof(struct param));
-	*param = zeroparam;
+	memset(param, 0, sizeof(struct param));
 	param->value = xmalloc(valuelen + 1);
-	strncpy(param->value, value, valuelen);
-	param->value[valuelen] = '\0';
+	strlcpy(param->value, value, valuelen + 1);
+
 	for (p = param->value; *p; p++) {
-	    if (islower(*p)) *p = toupper(*p);
+	    if (islower((int) *p)) *p = toupper((int) *p);
 	}
 
 	/* Get ready to parse the next parameter */
@@ -1258,31 +1255,33 @@ struct param **paramp;
  * Parse a RFC-822 date from a header.
  * Only parses to day granularity--ignores the time of day.
  */
-static time_t
-message_parse_date(hdr)
+time_t
+message_parse_date(hdr, flags)
 char *hdr;
+unsigned flags;
 {
     struct tm tm;
-    static struct tm zerotm;
+    time_t t;
     char month[4];
     static char *monthname[] = {
 	"jan", "feb", "mar", "apr", "may", "jun",
 	"jul", "aug", "sep", "oct", "nov", "dec"
     };
+    int zone_off = 0;
 
     if (!hdr) goto baddate;
 
-    tm = zerotm;
+    memset(&tm, 0, sizeof(tm));
 
     message_parse_rfc822space(&hdr);
     if (!hdr) goto baddate;
 
-    if (isalpha(*hdr)) {
+    if (isalpha((int) *hdr)) {
 	/* Day name -- skip over it */
 	hdr++;
-	if (!isalpha(*hdr)) goto baddate;
+	if (!isalpha((int) *hdr)) goto baddate;
 	hdr++;
-	if (!isalpha(*hdr)) goto baddate;
+	if (!isalpha((int) *hdr)) goto baddate;
 	hdr++;
 	message_parse_rfc822space(&hdr);
 	if (!hdr || *hdr++ != ',') goto baddate;
@@ -1290,9 +1289,9 @@ char *hdr;
 	if (!hdr) goto baddate;
     }
 
-    if (!isdigit(*hdr)) goto baddate;
+    if (!isdigit((int) *hdr)) goto baddate;
     tm.tm_mday = *hdr++ - '0';
-    if (isdigit(*hdr)) {
+    if (isdigit((int) *hdr)) {
 	tm.tm_mday = tm.tm_mday*10 + *hdr++ - '0';
     }
     
@@ -1300,11 +1299,11 @@ char *hdr;
     message_parse_rfc822space(&hdr);
     if (!hdr) goto baddate;
     month[0] = *hdr++;
-    if (!isalpha(month[0])) goto baddate;
+    if (!isalpha((int) month[0])) goto baddate;
     month[1] = *hdr++;
-    if (!isalpha(month[1])) goto baddate;
+    if (!isalpha((int) month[1])) goto baddate;
     month[2] = *hdr++;
-    if (!isalpha(month[2])) goto baddate;
+    if (!isalpha((int) month[2])) goto baddate;
     month[3] = '\0';
     lcase(month);
     for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++) {
@@ -1314,23 +1313,136 @@ char *hdr;
     
     /* Parse year */
     message_parse_rfc822space(&hdr);
-    if (!hdr || !isdigit(*hdr)) goto baddate;
+    if (!hdr || !isdigit((int) *hdr)) goto baddate;
     tm.tm_year = *hdr++ - '0';
-    if (!isdigit(*hdr)) goto baddate;
+    if (!isdigit((int) *hdr)) goto baddate;
     tm.tm_year = tm.tm_year * 10 + *hdr++ - '0';
-    if (isdigit(*hdr)) {
+    if (isdigit((int) *hdr)) {
 	if (tm.tm_year < 19) goto baddate;
 	tm.tm_year -= 19;
 	tm.tm_year = tm.tm_year * 10 + *hdr++ - '0';
-	if (!isdigit(*hdr)) goto baddate;
+	if (!isdigit((int) *hdr)) goto baddate;
 	tm.tm_year = tm.tm_year * 10 + *hdr++ - '0';
+    } else {
+	if (tm.tm_year < 70) {
+	    /* two-digit year, probably after 2000.
+	     * This patent was overturned, right?
+	     */
+	    tm.tm_year += 100;
+	}
     }
+    if (isdigit((int) *hdr)) {
+       /* five-digit date */
+       goto baddate;
+     }
+
+    message_parse_rfc822space(&hdr);
+    if (hdr && (flags & PARSE_TIME)) {
+	/* Parse hour */
+	if (!hdr || !isdigit((int) *hdr)) goto badtime;
+	tm.tm_hour = *hdr++ - '0';
+	if (!isdigit((int) *hdr)) goto badtime;
+	tm.tm_hour = tm.tm_hour * 10 + *hdr++ - '0';
+	if (!hdr || *hdr++ != ':') goto badtime;
+
+	/* Parse min */
+	if (!hdr || !isdigit((int) *hdr)) goto badtime;
+	tm.tm_min = *hdr++ - '0';
+	if (!isdigit((int) *hdr)) goto badtime;
+	tm.tm_min = tm.tm_min * 10 + *hdr++ - '0';
+
+	if (*hdr == ':') {
+	    /* Parse sec */
+	    if (!++hdr || !isdigit((int) *hdr)) goto badtime;
+	    tm.tm_sec = *hdr++ - '0';
+	    if (!isdigit((int) *hdr)) goto badtime;
+	    tm.tm_sec = tm.tm_sec * 10 + *hdr++ - '0';
+	}
+
+	message_parse_rfc822space(&hdr);
+	if (hdr && (flags & PARSE_ZONE)) {
+	    /* Parse timezone offset */
+	    if (*hdr == '+' || *hdr == '-') {
+		/* Parse numeric offset */
+		int east = (*hdr++ == '-');
+
+		if (!hdr || !isdigit((int) *hdr)) goto badzone;
+		zone_off = *hdr++ - '0';
+		if (!hdr || !isdigit((int) *hdr)) goto badzone;
+		zone_off = zone_off * 10 + *hdr++ - '0';
+		if (!hdr || !isdigit((int) *hdr)) goto badzone;
+		zone_off = zone_off * 6 + *hdr++ - '0';
+		if (!hdr || !isdigit((int) *hdr)) goto badzone;
+		zone_off = zone_off * 10 + *hdr++ - '0';
+
+		if (east) zone_off = -zone_off;
+	    }
+	    else if (isalpha((unsigned char) *hdr)) {
+		char zone[4];
+
+		zone[0] = *hdr++;
+		if (!isalpha((unsigned char) *hdr)) {
+		    /* Parse military (single-char) zone */
+		    zone[1] = '\0';
+		    lcase(zone);
+		    if (zone[0] < 'j')
+			zone_off = (zone[0] - 'a' + 1) * 60;
+		    else if (zone[0] == 'j')
+			goto badzone;
+		    else if (zone[0] <= 'm')
+			zone_off = (zone[0] - 'a') * 60;
+		    else if (zone[0] < 'z')
+			zone_off = ('m' - zone[0]) * 60;
+		    else
+			zone_off = 0;
+		}
+		else {
+		    zone[1] = *hdr++;
+		    if (!isalpha((unsigned char) *hdr)) {
+			/* Parse UT (universal time) */
+			zone[2] = '\0';
+			lcase(zone);
+			if (strcmp(zone, "ut")) goto badzone;
+			zone_off = 0;
+		    }
+		    else {
+			/* Parse 3-char time zone */
+			char *p;
+
+			zone[2] = *hdr;
+			zone[3] = '\0';
+			lcase(zone);
+			/* GMT (Greenwich mean time) */
+			if (!strcmp(zone, "gmt")) zone_off = 0;
+
+			/* US time zone */
+			else {
+			    p = strchr("aecmpyhb", zone[0]);
+			    if (!p || zone[2] != 't') goto badzone;
+			    zone_off = (strlen(p) - 12) * 60;
+			    if (zone[1] == 'd') zone_off += 60;
+			    else if (zone[1] != 's') goto badzone;
+			}
+		    }
+		}
+	    }
+	    else
+ badzone:
+		zone_off = 0;
+	}
+    }
+    else
+ badtime:
+	tm.tm_hour = 12;
 
     tm.tm_isdst = -1;
-    tm.tm_hour = 12;
-
-    return mktime(&tm);
 
+    t = mktime(&tm);
+    /* Don't return -1; it's never right.  Return the current time instead.
+     * That's much closer than 1969.
+     */
+    if (t >= 0) return (t - zone_off * 60);
+    
  baddate:
     return time(0);
 }
@@ -1346,7 +1458,7 @@ char **s;
     int commentlevel = 0;
 
     if (!p) return;
-    while (*p && (isspace(*p) || *p == '(')) {
+    while (*p && (isspace((int) *p) || *p == '(')) {
 	if (*p == '\n') {
 	    p++;
 	    if (*p != ' ' && *p != '\t') {
@@ -1403,12 +1515,12 @@ struct body *body;
 struct boundary *boundaries;
 {
     struct body preamble, epilogue;
-    static struct body zerobody;
     struct param *boundary;
     char *defaultContentType = DEFAULT_CONTENT_TYPE;
     int i, depth;
 
-    preamble = epilogue = zerobody;
+    memset(&preamble, 0, sizeof(struct body));
+    memset(&epilogue, 0, sizeof(struct body));
     if (strcmp(body->subtype, "DIGEST") == 0) {
 	defaultContentType = "MESSAGE/RFC822";
     }
@@ -1440,11 +1552,17 @@ struct boundary *boundaries;
     message_parse_content(msg, format, &preamble, boundaries);
 
     /* Parse the component body-parts */
-    while (boundaries->count == depth && msg->offset < msg->len) {
+    while (boundaries->count == depth) {
 	body->subpart = (struct body *)xrealloc((char *)body->subpart,
 				 (body->numparts+1)*sizeof(struct body));
 	message_parse_body(msg, format, &body->subpart[body->numparts++],
 			   defaultContentType, boundaries);
+	if (msg->offset == msg->len &&
+	    body->subpart[body->numparts-1].boundary_size == 0) {
+	    /* hit the end of the message, therefore end all pending
+	       multiparts */
+	    boundaries->count = 0;
+	}
     }
 
     if (boundaries->count == depth-1) {
@@ -1524,8 +1642,7 @@ struct boundary *boundaries;
 
 	if (line[0] == '-' && line[1] == '-' &&
 	    message_pendingboundary(line, boundaries->id, &boundaries->count)) {
-	    body->boundary_size = len +
-		(format == MAILBOX_FORMAT_NETNEWS && endline[-1] == '\n');
+	    body->boundary_size = len;
 	    body->boundary_lines++;
 	    if (body->content_lines) {
 		body->content_lines--;
@@ -1542,7 +1659,6 @@ struct boundary *boundaries;
 
 	if (endline[-1] == '\n') {
 	    body->content_lines++;
-	    if (format == MAILBOX_FORMAT_NETNEWS) body->content_size++;
 	}
     }
 }
@@ -1583,6 +1699,7 @@ char **boundaries;
 int *boundaryct;
 {
     int i, len;
+    int rfc2046_strict = config_getswitch(IMAPOPT_RFC2046_STRICT);
 
     if (s[0] != '-' || s[1] != '-') return(0);
     s+=2;
@@ -1591,6 +1708,16 @@ int *boundaryct;
 	len = strlen(boundaries[i]);
         if (!strncmp(s, boundaries[i], len)) {
             if (s[len] == '-' && s[len+1] == '-') *boundaryct = i;
+	    else if (!rfc2046_strict && s[len] && !isspace((int) s[len])) {
+		/* Allow substring matches in the boundary.
+		 *
+		 * If rfc2046_strict is enabled, boundaries containing
+		 * other boundaries as substrings will be treated as identical
+		 * (per RFC 2046 section 5.1.1).  Note that this will
+		 * break some messages created by Eudora 5.1 (and earlier).
+		 */
+		continue;
+	    }
             return(1);
         }
     }
@@ -1673,7 +1800,7 @@ struct body *body;
 }
 
 /* Append character 'c' to 'ibuf' */
-#define PUTIBUF(ibuf,c) (((ibuf)->end<(ibuf)->last || message_ibuf_ensure((ibuf),1)),(*((ibuf)->end)++ = (c)))
+#define PUTIBUF(ibuf,c) (((void)((ibuf)->end<(ibuf)->last || message_ibuf_ensure((ibuf),1))),(*((ibuf)->end)++ = (c)))
 
 /*
  * Write the IMAP envelope for 'body' to 'ibuf'
@@ -1742,13 +1869,13 @@ int newformat;
 
 	if (newformat) {
 	    PUTIBUF(ibuf, ' ');
-	    if (param = body->params) {
+	    if ((param = body->params)!=NULL) {
 		PUTIBUF(ibuf, '(');
 		while (param) {
 		    message_write_nstring(ibuf, param->attribute);
 		    PUTIBUF(ibuf, ' ');
 		    message_write_nstring(ibuf, param->value);
-		    if (param = param->next) {
+		    if ((param = param->next)!=NULL) {
 			PUTIBUF(ibuf, ' ');
 		    }
 		}
@@ -1760,13 +1887,13 @@ int newformat;
 		PUTIBUF(ibuf, '(');
 		message_write_nstring(ibuf, body->disposition);
 		PUTIBUF(ibuf, ' ');
-		if (param = body->disposition_params) {
+		if ((param = body->disposition_params)!=NULL) {
 		    PUTIBUF(ibuf, '(');
 		    while (param) {
 			message_write_nstring(ibuf, param->attribute);
 			PUTIBUF(ibuf, ' ');
 			message_write_nstring(ibuf, param->value);
-			if (param = param->next) {
+			if ((param = param->next)!=NULL) {
 			    PUTIBUF(ibuf, ' ');
 			}
 		    }
@@ -1779,11 +1906,11 @@ int newformat;
 		message_write_nstring(ibuf, (char *)0);
 	    }
 	    PUTIBUF(ibuf, ' ');
-	    if (param = body->language) {
+	    if ((param = body->language)!=NULL) {
 		PUTIBUF(ibuf, '(');
 		while (param) {
 		    message_write_nstring(ibuf, param->value);
-		    if (param = param->next) {
+		    if ((param = param->next)!=NULL) {
 			PUTIBUF(ibuf, ' ');
 		    }
 		}
@@ -1802,13 +1929,13 @@ int newformat;
     message_write_nstring(ibuf, body->subtype);
     PUTIBUF(ibuf, ' ');
 
-    if (param = body->params) {
+    if ((param = body->params)!=NULL) {
 	PUTIBUF(ibuf, '(');
 	while (param) {
 	    message_write_nstring(ibuf, param->attribute);
 	    PUTIBUF(ibuf, ' ');
 	    message_write_nstring(ibuf, param->value);
-	    if (param = param->next) {
+	    if ((param = param->next)!=NULL) {
 		PUTIBUF(ibuf, ' ');
 	    }
 	}
@@ -1850,13 +1977,13 @@ int newformat;
 	    PUTIBUF(ibuf, '(');
 	    message_write_nstring(ibuf, body->disposition);
 	    PUTIBUF(ibuf, ' ');
-	    if (param = body->disposition_params) {
+	    if ((param = body->disposition_params)!=NULL) {
 		PUTIBUF(ibuf, '(');
 		while (param) {
 		    message_write_nstring(ibuf, param->attribute);
 		    PUTIBUF(ibuf, ' ');
 		    message_write_nstring(ibuf, param->value);
-		    if (param = param->next) {
+		    if ((param = param->next)!=NULL) {
 			PUTIBUF(ibuf, ' ');
 		    }
 		}
@@ -1869,11 +1996,11 @@ int newformat;
 	    message_write_nstring(ibuf, (char *)0);
 	}
 	PUTIBUF(ibuf, ' ');
-	if (param = body->language) {
+	if ((param = body->language)!=NULL) {
 	    PUTIBUF(ibuf, '(');
 	    while (param) {
 		message_write_nstring(ibuf, param->value);
-		if (param = param->next) {
+		if ((param = param->next)!=NULL) {
 		    PUTIBUF(ibuf, ' ');
 		}
 	    }
@@ -1947,7 +2074,7 @@ char *s;
     if (*p || len >= 1024) {
 	/* Write out as literal */
 	char buf[100];
-	sprintf(buf, "{%u}\r\n", strlen(s));
+	snprintf(buf, sizeof(buf), "{%u}\r\n", strlen(s));
 	message_ibuf_ensure(ibuf, strlen(s)+strlen(buf));
 	for (p = buf; *p; p++) *(ibuf->end)++ = *p;
 	for (p = s; *p; p++) *(ibuf->end)++ = *p;
@@ -1985,7 +2112,7 @@ unsigned n;
 {
     char buf[100], *p;
 
-    sprintf(buf, "%u", n);
+    snprintf(buf, sizeof(buf), "%u", n);
 
     message_ibuf_ensure(ibuf, strlen(buf));
     for (p = buf; *p; p++) *(ibuf->end)++ = *p;
@@ -2259,15 +2386,14 @@ struct ibuf *ibuf;
 /*
  * Ensure 'ibuf' has enough free space to append 'len' bytes.
  */
-static
-message_ibuf_ensure(ibuf, len)
-struct ibuf *ibuf;
-unsigned len;
+static int
+message_ibuf_ensure(struct ibuf *ibuf,
+		    unsigned len)
 {
     char *s;
     int size;
 
-    if (ibuf->last - ibuf->end >= len) return;
+    if (ibuf->last - ibuf->end >= len) return 0;
     if (len < IBUFGROWSIZE) len = IBUFGROWSIZE;
 
     s = ibuf->start - sizeof(bit32);
@@ -2277,6 +2403,8 @@ unsigned len;
     ibuf->end = (ibuf->end - ibuf->start) + s;
     ibuf->start = s;
     ibuf->last = s + size;
+
+    return 1;
 }
 
 /*
@@ -2289,7 +2417,6 @@ struct ibuf *ibuf;
 {
     char *s;
     int len;
-    int pad;
 
     /* Drop padding on end */
     message_ibuf_ensure(ibuf, 3);
