@@ -1,18 +1,48 @@
 /* mkchartable.c -- Generate character set mapping table
  *
- * Copyright 1996, Carnegie Mellon University.  All Rights Reserved.
- * 
- * This software is made available for academic and research
- * purposes only.  No commercial license is hereby granted.
- * Copying and other reproduction is authorized only for research,
- * education, and other non-commercial purposes.  No warranties,
- * either expressed or implied, are made regarding the operation,
- * use, or results of the software.  Such a release does not permit
- * use of the code for commercial purposes or benefits by anyone
- * without specific, additional permission by the owner of the code.
+ * $Id: mkchartable.c,v 1.19.16.5 2003/05/14 19:43:06 ken3 Exp $
+ *
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
  */
 
+#include <config.h>
 #include <ctype.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -157,10 +187,16 @@ main(int argc, char **argv)
     printf("    { \"iso-2022-kr\", chartables_iso_2022_kr },\n");
     printf("    { \"gb2312\", chartables_gb2312 },\n");
     printf("    { \"big5\", chartables_big5 },\n");
-    printf("    /* Compatiblilty names */\n");
+    printf("    /* Compatibility names */\n");
     printf("    { \"unicode-1-1-utf-7\", chartables_utf_7 },\n");
     printf("    { \"unicode-2-0-utf-7\", chartables_utf_7 },\n");
     printf("    { \"x-unicode-2-0-utf-7\", chartables_utf_7 },\n");
+    printf("    /* End Compatibility Names */\n");
+    printf("    { \"iso-8859-15\", chartables_iso_8859_15 },\n");
+    printf("    { \"windows-1252\", chartables_windows_1252 },\n");
+    printf("    { \"windows-1256\", chartables_windows_1256 },\n");
+    printf("    /* New character sets should only be added to end so that\n");
+    printf("     * cache files stay with valid information */\n");
     printf("};\n");
     printf("const int chartables_num_charsets = (sizeof(chartables_charset_table)/sizeof(*chartables_charset_table));\n");
 
@@ -437,7 +473,9 @@ mungemappings(void)
 		printf(" END, /* Translation for %04x (offset %04x) */\n",
 		       map[n].code, last_translation);
 		map[n].trans_offset = last_translation;
-		last_translation += n_t;
+
+                /* last_translation points to the offset the next translation will start from */
+		last_translation += n_t + 1;
 	    }
 	}
     }
@@ -516,7 +554,12 @@ readcharfile(char *name)
 	if (i > 4) goto syntaxerr;	
 	if (i > 2) {
 	    if (EMPTYTCHAR(table[thisstate].ch[thischar>>8])) {
+                /* we create a new state (not in the input file) to
+                   deal with multibyte characters that start with the
+                   byte 'thischar >> 8'. */
+
 		char action[1024];
+
 		sprintf(action, ">%s_%02x <", table[thisstate].name,
 			thischar>>8);
 		table[thisstate].ch[thischar>>8].action = xstrdup(action);
@@ -526,12 +569,21 @@ readcharfile(char *name)
 	    }
 	    else if (!table[thisstate].ch[thischar>>8].action ||
 		     table[thisstate].ch[thischar>>8].action[0] != '>') {
+                /* either we think this byte isn't the start of a
+                   multibyte character, or the action associated with this
+                   byte isn't a state change. */
+
 		fprintf(stderr,
 			"%s: line %d: multibyte/single-byte conflict\n",
 			name, line);
 		exit(1);
 	    }
 	    else {
+                /* we find the already created state to deal with multibytes
+                   starting with 'thischar >> 8' and move to it so we
+                   insert the 2nd byte of this multibyte char in the right
+                   state. */
+
 		thisstate =
 		  findstate(table[thisstate].ch[thischar>>8].action+1);
 		if (thisstate == -1) {
@@ -659,7 +711,9 @@ static void mkutf8table(void)
 	table[thisstate].ch[thischar].action = "U83_3";
     }
 
-    /* Populate 2-char sequences */
+    /* Populate 2-char sequences---the first byte shifts to another
+     * state; the 2nd byte chooses the character, just like any other
+     * 2-byte encoding */
     for (prefix = 2; prefix <= 0x1f; prefix++) {
 	sprintf(buf, ">STATE-2-%02x", prefix);
 	table[start_state].ch[prefix+0xc0].action = xstrdup(buf);
@@ -670,7 +724,8 @@ static void mkutf8table(void)
 	}
     }
 
-    /* Populate 3-char sequences */
+    /* Populate 3-char sequences, which the decoder handles
+     * magically, outside of the state system. */
     for (thischar = 0xe0; thischar <= 0xef; thischar++) {
 	table[start_state].ch[thischar].action = "U83";
     }
@@ -726,7 +781,7 @@ newstate(char *args)
 
     if (table_num == table_alloc) {
 	table_alloc += TABLEGROW;
-	table = (struct table *)realloc((char *)table,
+	table = (struct table *)xrealloc((char *)table,
 					 table_alloc * sizeof(struct table));
     }
 
